@@ -1,5 +1,6 @@
+{-# LANGUAGE NoImplicitPrelude #-}
 {-
-Copyright (C) 2008-2016 John MacFarlane <jgm@berkeley.edu>
+Copyright (C) 2008-2018 John MacFarlane <jgm@berkeley.edu>
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -18,7 +19,7 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 {- |
    Module      : Text.Pandoc.Highlighting
-   Copyright   : Copyright (C) 2008-2016 John MacFarlane
+   Copyright   : Copyright (C) 2008-2018 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -28,7 +29,8 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 Exports functions for syntax highlighting.
 -}
 
-module Text.Pandoc.Highlighting ( languages
+module Text.Pandoc.Highlighting ( highlightingStyles
+                                , languages
                                 , languagesByExtension
                                 , highlight
                                 , formatLaTeXInline
@@ -48,14 +50,26 @@ module Text.Pandoc.Highlighting ( languages
                                 , fromListingsLanguage
                                 , toListingsLanguage
                                 ) where
-import Text.Pandoc.Definition
-import Text.Pandoc.Shared (safeRead)
-import Skylighting
-import Data.Maybe (fromMaybe)
+import Prelude
+import Control.Monad
 import Data.Char (toLower)
 import qualified Data.Map as M
-import Control.Monad
+import Data.Maybe (fromMaybe)
 import qualified Data.Text as T
+import Skylighting
+import Text.Pandoc.Definition
+import Text.Pandoc.Shared (safeRead)
+
+highlightingStyles :: [(String, Style)]
+highlightingStyles =
+  [("pygments", pygments),
+   ("tango", tango),
+   ("espresso", espresso),
+   ("zenburn", zenburn),
+   ("kate", kate),
+   ("monochrome", monochrome),
+   ("breezedark", breezeDark),
+   ("haddock", haddock)]
 
 languages :: [String]
 languages = [T.unpack (T.toLower (sName s)) | s <- M.elems defaultSyntaxMap]
@@ -64,34 +78,39 @@ languagesByExtension :: String -> [String]
 languagesByExtension ext =
   [T.unpack (T.toLower (sName s)) | s <- syntaxesByExtension defaultSyntaxMap ext]
 
-highlight :: (FormatOptions -> [SourceLine] -> a) -- ^ Formatter
+highlight :: SyntaxMap
+          -> (FormatOptions -> [SourceLine] -> a) -- ^ Formatter
           -> Attr   -- ^ Attributes of the CodeBlock
           -> String -- ^ Raw contents of the CodeBlock
-          -> Maybe a -- ^ Maybe the formatted result
-highlight formatter (_, classes, keyvals) rawCode =
+          -> Either String a
+highlight syntaxmap formatter (ident, classes, keyvals) rawCode =
   let firstNum = fromMaybe 1 (safeRead (fromMaybe "1" $ lookup "startFrom" keyvals))
       fmtOpts = defaultFormatOpts{
                   startNumber = firstNum,
+                  lineAnchors = any (`elem`
+                        ["line-anchors", "lineAnchors"]) classes,
                   numberLines = any (`elem`
-                        ["number","numberLines", "number-lines"]) classes }
-      tokenizeOpts = TokenizerConfig{ syntaxMap = defaultSyntaxMap
+                        ["number","numberLines", "number-lines"]) classes,
+                  lineIdPrefix = if null ident
+                                    then mempty
+                                    else T.pack (ident ++ "-") }
+      tokenizeOpts = TokenizerConfig{ syntaxMap = syntaxmap
                                     , traceOutput = False }
       classes' = map T.pack classes
       rawCode' = T.pack rawCode
-  in  case msum (map (\l -> lookupSyntax l defaultSyntaxMap) classes') of
+  in  case msum (map (`lookupSyntax` syntaxmap) classes') of
             Nothing
-              | numberLines fmtOpts -> Just
+              | numberLines fmtOpts -> Right
                               $ formatter fmtOpts{ codeClasses = [],
                                                    containerClasses = classes' }
-                              $ map (\ln -> [(NormalTok, ln)]) $ T.lines rawCode'
-              | otherwise  -> Nothing
+                              $ map (\ln -> [(NormalTok, ln)])
+                              $ T.lines rawCode'
+              | otherwise  -> Left ""
             Just syntax  ->
-              case tokenize tokenizeOpts syntax rawCode' of
-                   Right slines -> Just $
-                         formatter fmtOpts{ codeClasses =
-                                               [T.toLower (sShortname syntax)],
-                                            containerClasses = classes' } slines
-                   Left _ -> Nothing
+              formatter fmtOpts{ codeClasses =
+                                   [T.toLower (sShortname syntax)],
+                                  containerClasses = classes' } <$>
+                tokenize tokenizeOpts syntax rawCode'
 
 -- Functions for correlating latex listings package's language names
 -- with skylighting language names:
