@@ -1,3 +1,4 @@
+{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE CPP #-}
 {- |
    Module      : Text.Pandoc.Readers.Haddock
@@ -14,33 +15,40 @@ module Text.Pandoc.Readers.Haddock
     ( readHaddock
     ) where
 
-import Text.Pandoc.Builder (Blocks, Inlines)
-import qualified Text.Pandoc.Builder as B
-import Data.Monoid ((<>))
-import Text.Pandoc.Shared (trim, splitBy)
+import Prelude
+import Control.Monad.Except (throwError)
 import Data.List (intersperse, stripPrefix)
 import Data.Maybe (fromMaybe)
-import Text.Pandoc.Definition
-import Text.Pandoc.Options
+import Data.Text (Text, unpack)
 import Documentation.Haddock.Parser
-import Documentation.Haddock.Types
-import Debug.Trace (trace)
-
+import Documentation.Haddock.Types as H
+import Text.Pandoc.Builder (Blocks, Inlines)
+import qualified Text.Pandoc.Builder as B
+import Text.Pandoc.Class (PandocMonad)
+import Text.Pandoc.Definition
 import Text.Pandoc.Error
+import Text.Pandoc.Options
+import Text.Pandoc.Shared (crFilter, splitBy, trim)
+
 
 -- | Parse Haddock markup and return a 'Pandoc' document.
-readHaddock :: ReaderOptions -- ^ Reader options
-            -> String        -- ^ String to parse
-            -> Either PandocError Pandoc
-readHaddock opts =
+readHaddock :: PandocMonad m
+            => ReaderOptions
+            -> Text
+            -> m Pandoc
+readHaddock opts s = case readHaddockEither opts (unpack (crFilter s)) of
+  Right result -> return result
+  Left e       -> throwError e
+
+readHaddockEither :: ReaderOptions -- ^ Reader options
+                  -> String        -- ^ String to parse
+                  -> Either PandocError Pandoc
+readHaddockEither _opts =
 #if MIN_VERSION_haddock_library(1,2,0)
-  Right . B.doc . docHToBlocks . trace' . _doc . parseParas
+  Right . B.doc . docHToBlocks . _doc . parseParas
 #else
-  Right .  B.doc . docHToBlocks . trace' . parseParas
+  Right .  B.doc . docHToBlocks . parseParas
 #endif
-  where trace' x = if readerTrace opts
-                      then trace (show x) x
-                      else x
 
 docHToBlocks :: DocH String Identifier -> Blocks
 docHToBlocks d' =
@@ -79,6 +87,20 @@ docHToBlocks d' =
     DocProperty s -> B.codeBlockWith ("",["property","haskell"],[]) (trim s)
     DocExamples es -> mconcat $ map (\e ->
        makeExample ">>>" (exampleExpression e) (exampleResult e)) es
+#if MIN_VERSION_haddock_library(1,5,0)
+    DocTable H.Table{ tableHeaderRows = headerRows
+                    , tableBodyRows = bodyRows
+                    }
+      -> let toCells = map (docHToBlocks . tableCellContents) . tableRowCells
+             (header, body) =
+               if null headerRows
+                  then ([], map toCells bodyRows)
+                  else (toCells (head headerRows),
+                        map toCells (tail headerRows ++ bodyRows))
+             colspecs = replicate (maximum (map length body))
+                             (AlignDefault, 0.0)
+         in  B.table mempty colspecs header body
+#endif
 
   where inlineFallback = B.plain $ docHToInlines False d'
         consolidatePlains = B.fromList . consolidatePlains' . B.toList
@@ -90,7 +112,7 @@ docHToBlocks d' =
         isPlain (Plain _) = True
         isPlain _         = False
         extractContents (Plain xs) = xs
-        extractContents _ = []
+        extractContents _          = []
 
 docHToInlines :: Bool -> DocH String Identifier -> Inlines
 docHToInlines isCode d' =
@@ -127,6 +149,9 @@ docHToInlines isCode d' =
     DocAName s -> B.spanWith (s,["anchor"],[]) mempty
     DocProperty _ -> mempty
     DocExamples _ -> mempty
+#if MIN_VERSION_haddock_library(1,5,0)
+    DocTable _ -> mempty
+#endif
 
 -- | Create an 'Example', stripping superfluous characters as appropriate
 makeExample :: String -> String -> [String] -> Blocks
@@ -135,7 +160,7 @@ makeExample prompt expression result =
         <> B.space
         <> B.codeWith ([], ["haskell","expr"], []) (trim expression)
         <> B.linebreak
-        <> (mconcat $ intersperse B.linebreak $ map coder result')
+        <> mconcat (intersperse B.linebreak $ map coder result')
   where
     -- 1. drop trailing whitespace from the prompt, remember the prefix
     prefix = takeWhile (`elem` " \t") prompt
