@@ -1,6 +1,7 @@
+{-# LANGUAGE NoImplicitPrelude #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-
-Copyright (C) 2014-2016 Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
+Copyright (C) 2014-2018 Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -18,8 +19,8 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 -}
 
 {- |
-   Module      : Text.Pandoc.Readers.Org.Options
-   Copyright   : Copyright (C) 2014-2016 Albert Krewinkel
+   Module      : Text.Pandoc.Readers.Org.Shared
+   Copyright   : Copyright (C) 2014-2018 Albert Krewinkel
    License     : GNU GPL, version 2 or above
 
    Maintainer  : Albert Krewinkel <tarleb+pandoc@moltkeplatz.de>
@@ -29,24 +30,24 @@ Utility functions used in other Pandoc Org modules.
 module Text.Pandoc.Readers.Org.Shared
   ( cleanLinkString
   , isImageFilename
-  , rundocBlockClass
-  , toRundocAttrib
+  , originalLang
   , translateLang
   ) where
 
-import           Control.Arrow ( first )
-import           Data.Char ( isAlphaNum )
-import           Data.List ( isPrefixOf, isSuffixOf )
+import Prelude
+import Data.Char (isAlphaNum)
+import Data.List (isPrefixOf)
+import System.FilePath (isValid, takeExtension)
 
 
 -- | Check whether the given string looks like the path to of URL of an image.
 isImageFilename :: String -> Bool
-isImageFilename filename =
-  any (\x -> ('.':x)  `isSuffixOf` filename) imageExtensions &&
-  (any (\x -> (x ++ "://") `isPrefixOf` filename) protocols ||
-   ':' `notElem` filename)
+isImageFilename fp = hasImageExtension && (isValid fp || isKnownProtocolUri)
  where
-   imageExtensions = [ "jpeg" , "jpg" , "png" , "gif" , "svg" ]
+   hasImageExtension = takeExtension fp `elem` imageExtensions
+   isKnownProtocolUri = any (\x -> (x ++ "://") `isPrefixOf` fp) protocols
+
+   imageExtensions = [ ".jpeg", ".jpg", ".png", ".gif", ".svg" ]
    protocols = [ "file", "http", "https" ]
 
 -- | Cleanup and canonicalize a string describing a link.  Return @Nothing@ if
@@ -58,8 +59,8 @@ cleanLinkString s =
     '.':'/':_              -> Just s                 -- relative path
     '.':'.':'/':_          -> Just s                 -- relative path
     -- Relative path or URL (file schema)
-    'f':'i':'l':'e':':':s' -> Just $ if ("//" `isPrefixOf` s') then s else s'
-    _ | isUrl s            -> Just s                 -- URL
+    'f':'i':'l':'e':':':s' -> Just $ if "//" `isPrefixOf` s' then s else s'
+    _                      | isUrl s            -> Just s                 -- URL
     _                      -> Nothing
  where
    isUrl :: String -> Bool
@@ -68,17 +69,17 @@ cleanLinkString s =
      in all (\c -> isAlphaNum c || c `elem` (".-"::String)) scheme
           && not (null path)
 
--- | Prefix used for Rundoc classes and arguments.
-rundocPrefix :: String
-rundocPrefix = "rundoc-"
+-- | Creates an key-value pair marking the original language name specified for
+-- a piece of source code.
 
--- | The class-name used to mark rundoc blocks.
-rundocBlockClass :: String
-rundocBlockClass = rundocPrefix ++ "block"
-
--- | Prefix the name of a attribute, marking it as a code execution parameter.
-toRundocAttrib :: (String, String) -> (String, String)
-toRundocAttrib = first (rundocPrefix ++)
+-- | Creates an key-value attributes marking the original language name
+-- specified for a piece of source code.
+originalLang :: String -> [(String, String)]
+originalLang lang =
+  let transLang = translateLang lang
+  in if transLang == lang
+     then []
+     else [("org-language", lang)]
 
 -- | Translate from Org-mode's programming language identifiers to those used
 -- by Pandoc.  This is useful to allow for proper syntax highlighting in
