@@ -1,8 +1,11 @@
-{-# LANGUAGE OverloadedStrings, FlexibleContexts #-}
+{-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 
 {- |
    Module      : Text.Pandoc.Writers.ICML
-   Copyright   : Copyright (C) 2013-2016 github.com/mb21
+   Copyright   : Copyright (C) 2013-2018 github.com/mb21
    License     : GNU GPL, version 2 or above
 
    Stability   : alpha
@@ -14,20 +17,26 @@ InCopy is the companion word-processor to Adobe InDesign and ICML documents can
 into InDesign with File -> Place.
 -}
 module Text.Pandoc.Writers.ICML (writeICML) where
-import Text.Pandoc.Definition
-import Text.Pandoc.XML
-import Text.Pandoc.Readers.TeXMath (texMathToInlines)
-import Text.Pandoc.Writers.Shared
-import Text.Pandoc.Shared (linesToPara, splitBy, fetchItem, warn)
-import Text.Pandoc.Options
-import Text.Pandoc.Templates (renderTemplate')
-import Text.Pandoc.Pretty
-import Text.Pandoc.ImageSize
-import Data.List (isPrefixOf, isInfixOf, stripPrefix, intersperse)
-import Data.Text as Text (breakOnAll, pack)
-import Control.Monad.State
-import Network.URI (isURI)
+import Prelude
+import Control.Monad.Except (catchError)
+import Control.Monad.State.Strict
+import Data.List (intersperse, isInfixOf, isPrefixOf, stripPrefix)
+import Data.Maybe (fromMaybe)
 import qualified Data.Set as Set
+import Data.Text as Text (breakOnAll, pack)
+import Data.Text (Text)
+import Text.Pandoc.Class (PandocMonad, report)
+import qualified Text.Pandoc.Class as P
+import Text.Pandoc.Definition
+import Text.Pandoc.ImageSize
+import Text.Pandoc.Logging
+import Text.Pandoc.Options
+import Text.Pandoc.Pretty
+import Text.Pandoc.Shared (isURI, linesToPara, splitBy)
+import Text.Pandoc.Templates (renderTemplate')
+import Text.Pandoc.Writers.Math (texMathToInlines)
+import Text.Pandoc.Writers.Shared
+import Text.Pandoc.XML
 
 type Style = [String]
 type Hyperlink = [(Int, String)]
@@ -40,7 +49,7 @@ data WriterState = WriterState{
   , maxListDepth :: Int
   }
 
-type WS a = StateT WriterState IO a
+type WS m = StateT WriterState m
 
 defaultWriterState :: WriterState
 defaultWriterState = WriterState{
@@ -121,13 +130,13 @@ subListParName    = "subParagraph"
 footnoteName      = "Footnote"
 citeName          = "Cite"
 
-
 -- | Convert Pandoc document to string in ICML format.
-writeICML :: WriterOptions -> Pandoc -> IO String
+writeICML :: PandocMonad m => WriterOptions -> Pandoc -> m Text
 writeICML opts (Pandoc meta blocks) = do
   let colwidth = if writerWrapText opts == WrapAuto
                     then Just $ writerColumns opts
                     else Nothing
+      render' :: Doc -> Text
       render' = render colwidth
       renderMeta f s = liftM (render' . fst) $ runStateT (f opts [] s) defaultWriterState
   metadata <- metaToJSON opts
@@ -139,18 +148,15 @@ writeICML opts (Pandoc meta blocks) = do
       context = defField "body" main
               $ defField "charStyles" (render' $ charStylesToDoc st)
               $ defField "parStyles"  (render' $ parStylesToDoc st)
-              $ defField "hyperlinks" (render' $ hyperlinksToDoc $ links st)
-              $ metadata
-  return $ case writerTemplate opts of
-                Nothing  -> main
-                Just tpl -> renderTemplate' tpl context
+              $ defField "hyperlinks" (render' $ hyperlinksToDoc $ links st) metadata
+  case writerTemplate opts of
+       Nothing  -> return main
+       Just tpl -> renderTemplate' tpl context
 
 -- | Auxilary functions for parStylesToDoc and charStylesToDoc.
 contains :: String -> (String, (String, String)) -> [(String, String)]
 contains s rule =
-  if isInfixOf (fst rule) s
-     then [snd rule]
-     else []
+  [snd rule | (fst rule) `isInfixOf` s]
 
 -- | The monospaced font to use as default.
 monospacedFont :: Doc
@@ -174,7 +180,7 @@ parStylesToDoc st = vcat $ map makeStyle $ Set.toAscList $ blockStyles st
   where
     makeStyle s =
       let countSubStrs sub str = length $ Text.breakOnAll (Text.pack sub) (Text.pack str)
-          attrs = concat $ map (contains s) $ [
+          attrs = concatMap (contains s) [
                                (defListTermName, ("BulletsAndNumberingListType", "BulletList"))
                              , (defListTermName, ("FontStyle", "Bold"))
                              , (tableHeaderName, ("FontStyle", "Bold"))
@@ -200,9 +206,9 @@ parStylesToDoc st = vcat $ map makeStyle $ Set.toAscList $ blockStyles st
             where
               numbering | isOrderedList = [("NumberingExpression", "^#.^t"), ("NumberingLevel", show nOrds)]
                         | otherwise     = []
-              listType | isOrderedList && (not $ isInfixOf subListParName s)
+              listType | isOrderedList && not (subListParName `isInfixOf` s)
                            = [("BulletsAndNumberingListType", "NumberedList")]
-                       | isBulletList && (not $ isInfixOf subListParName s)
+                       | isBulletList && not (subListParName `isInfixOf` s)
                            = [("BulletsAndNumberingListType", "BulletList")]
                        | otherwise = []
               indent = [("LeftIndent", show indt)]
@@ -210,9 +216,9 @@ parStylesToDoc st = vcat $ map makeStyle $ Set.toAscList $ blockStyles st
                   nBlockQuotes = countSubStrs blockQuoteName s
                   nDefLists = countSubStrs defListDefName s
                   indt = max 0 $ defaultListIndent*(nBuls + nOrds - 1) + defaultIndent*(nBlockQuotes + nDefLists)
-          props = inTags True "Properties" [] $ (basedOn $$ tabList $$ numbForm)
+          props = inTags True "Properties" [] (basedOn $$ tabList $$ numbForm)
             where
-              font = if isInfixOf codeBlockName s
+              font = if codeBlockName `isInfixOf` s
                         then monospacedFont
                         else empty
               basedOn = inTags False "BasedOn" [("type", "object")] (text "$ID/NormalParagraphStyle") $$ font
@@ -239,7 +245,7 @@ charStylesToDoc :: WriterState -> Doc
 charStylesToDoc st = vcat $ map makeStyle $ Set.toAscList $ inlineStyles st
   where
     makeStyle s =
-      let attrs = concat $ map (contains s) [
+      let attrs = concatMap (contains s) [
                                (strikeoutName,   ("StrikeThru", "true"))
                              , (superscriptName, ("Position", "Superscript"))
                              , (subscriptName,   ("Position", "Subscript"))
@@ -253,7 +259,7 @@ charStylesToDoc st = vcat $ map makeStyle $ Set.toAscList $ inlineStyles st
                     inTags False "BasedOn" [("type", "object")] (text "$ID/NormalCharacterStyle") $$ font
                   where
                     font =
-                      if isInfixOf codeName s
+                      if codeName `isInfixOf` s
                          then monospacedFont
                          else empty
       in  inTags True "CharacterStyle" ([("Self", "CharacterStyle/"++s), ("Name", s)] ++ attrs') props
@@ -273,23 +279,22 @@ hyperlinksToDoc (x:xs) = hyp x $$ hyperlinksToDoc xs
     hyp (ident, url) = hdest $$ hlink
       where
         hdest = selfClosingTag "HyperlinkURLDestination"
-                  [("Self", "HyperlinkURLDestination/"++(escapeColons url)), ("Name","link"), ("DestinationURL",url), ("DestinationUniqueKey","1")] -- HyperlinkURLDestination with more than one colon crashes CS6
+                  [("Self", "HyperlinkURLDestination/"++escapeColons url), ("Name","link"), ("DestinationURL",url), ("DestinationUniqueKey","1")] -- HyperlinkURLDestination with more than one colon crashes CS6
         hlink = inTags True "Hyperlink" [("Self","uf-"++show ident),  ("Name",url),
                     ("Source","htss-"++show ident), ("Visible","true"), ("DestinationUniqueKey","1")]
                   $ inTags True "Properties" []
                   $ inTags False "BorderColor" [("type","enumeration")] (text "Black")
-                  $$ (inTags False "Destination" [("type","object")]
-                  $ text $ "HyperlinkURLDestination/"++(escapeColons (escapeStringForXML url))) -- HyperlinkURLDestination with more than one colon crashes CS6
+                  $$ inTags False "Destination" [("type","object")] (text $ "HyperlinkURLDestination/"++escapeColons (escapeStringForXML url)) -- HyperlinkURLDestination with more than one colon crashes CS6
 
 
 -- | Convert a list of Pandoc blocks to ICML.
-blocksToICML :: WriterOptions -> Style -> [Block] -> WS Doc
+blocksToICML :: PandocMonad m => WriterOptions -> Style -> [Block] -> WS m Doc
 blocksToICML opts style lst = do
   docs <- mapM (blockToICML opts style) lst
   return $ intersperseBrs docs
 
 -- | Convert a Pandoc block element to ICML.
-blockToICML :: WriterOptions -> Style -> Block -> WS Doc
+blockToICML :: PandocMonad m => WriterOptions -> Style -> Block -> WS m Doc
 blockToICML opts style (Plain lst) = parStyle opts style lst
 -- title beginning with fig: indicates that the image is a figure
 blockToICML opts style (Para img@[Image _ txt (_,'f':'i':'g':':':_)]) = do
@@ -299,10 +304,12 @@ blockToICML opts style (Para img@[Image _ txt (_,'f':'i':'g':':':_)]) = do
 blockToICML opts style (Para lst) = parStyle opts (paragraphName:style) lst
 blockToICML opts style (LineBlock lns) =
   blockToICML opts style $ linesToPara lns
-blockToICML opts style (CodeBlock _ str) = parStyle opts (codeBlockName:style) $ [Str str]
-blockToICML _ _ (RawBlock f str)
+blockToICML opts style (CodeBlock _ str) = parStyle opts (codeBlockName:style) [Str str]
+blockToICML _ _ b@(RawBlock f str)
   | f == Format "icml" = return $ text str
-  | otherwise          = return empty
+  | otherwise          = do
+      report $ BlockNotRendered b
+      return empty
 blockToICML opts style (BlockQuote blocks) = blocksToICML opts (blockQuoteName:style) blocks
 blockToICML opts style (OrderedList attribs lst) = listItemsToICML opts orderedListName style (Just attribs) lst
 blockToICML opts style (BulletList lst) = listItemsToICML opts bulletListName style Nothing lst
@@ -343,11 +350,10 @@ blockToICML opts style (Table caption aligns widths headers rows) =
                     then rows
                     else headers:rows
       cells <- rowsToICML tabl (0::Int)
-      let colWidths w = if w > 0
-                           then [("SingleColumnWidth",show $ 500 * w)]
-                           else []
-      let tupToDoc tup = selfClosingTag "Column" $ [("Name",show $ fst tup)] ++ (colWidths $ snd tup)
-      let colDescs = vcat $ map tupToDoc $ zip [0..nrCols-1] widths
+      let colWidths w =
+            [("SingleColumnWidth",show $ 500 * w) | w > 0]
+      let tupToDoc tup = selfClosingTag "Column" $ ("Name",show $ fst tup) : colWidths (snd tup)
+      let colDescs = vcat $ zipWith (curry tupToDoc) [0..nrCols-1] widths
       let tableDoc = return $ inTags True "Table" [
                          ("AppliedTableStyle","TableStyle/Table")
                        , ("HeaderRowCount", nrHeaders)
@@ -359,7 +365,7 @@ blockToICML opts style (Div _ lst) = blocksToICML opts style lst
 blockToICML _ _ Null = return empty
 
 -- | Convert a list of lists of blocks to ICML list items.
-listItemsToICML :: WriterOptions -> String -> Style -> Maybe ListAttributes -> [[Block]] -> WS Doc
+listItemsToICML :: PandocMonad m => WriterOptions -> String -> Style -> Maybe ListAttributes -> [[Block]] -> WS m Doc
 listItemsToICML _ _ _ _ [] = return empty
 listItemsToICML opts listType style attribs (first:rest) = do
   st <- get
@@ -374,18 +380,17 @@ listItemsToICML opts listType style attribs (first:rest) = do
   return $ intersperseBrs docs
 
 -- | Convert a list of blocks to ICML list items.
-listItemToICML :: WriterOptions -> Style -> Bool-> Maybe ListAttributes -> [Block] -> WS Doc
+listItemToICML :: PandocMonad m => WriterOptions -> Style -> Bool-> Maybe ListAttributes -> [Block] -> WS m Doc
 listItemToICML opts style isFirst attribs item =
   let makeNumbStart (Just (beginsWith, numbStl, _)) =
         let doN DefaultStyle = []
-            doN LowerRoman = [lowerRomanName]
-            doN UpperRoman = [upperRomanName]
-            doN LowerAlpha = [lowerAlphaName]
-            doN UpperAlpha = [upperAlphaName]
-            doN _ = []
-            bw = if beginsWith > 1
-                    then [beginsWithName ++ show beginsWith]
-                    else []
+            doN LowerRoman   = [lowerRomanName]
+            doN UpperRoman   = [upperRomanName]
+            doN LowerAlpha   = [lowerAlphaName]
+            doN UpperAlpha   = [upperAlphaName]
+            doN _            = []
+            bw =
+              [beginsWithName ++ show beginsWith | beginsWith > 1]
         in  doN numbStl ++ bw
       makeNumbStart Nothing = []
       stl = if isFirst
@@ -394,26 +399,26 @@ listItemToICML opts style isFirst attribs item =
       stl' = makeNumbStart attribs ++ stl
   in  if length item > 1
          then do
-           let insertTab (Para lst) = blockToICML opts (subListParName:style) $ Para $ (Str "\t"):lst
+           let insertTab (Para lst) = blockToICML opts (subListParName:style) $ Para $ Str "\t":lst
                insertTab block      = blockToICML opts style block
            f <- blockToICML opts stl' $ head item
            r <- mapM insertTab $ tail item
            return $ intersperseBrs (f : r)
          else blocksToICML opts stl' item
 
-definitionListItemToICML :: WriterOptions -> Style -> ([Inline],[[Block]]) -> WS Doc
+definitionListItemToICML :: PandocMonad m => WriterOptions -> Style -> ([Inline],[[Block]]) -> WS m Doc
 definitionListItemToICML opts style (term,defs) = do
   term' <- parStyle opts (defListTermName:style) term
   defs' <- mapM (blocksToICML opts (defListDefName:style)) defs
-  return $ intersperseBrs $ (term' : defs')
+  return $ intersperseBrs (term' : defs')
 
 
 -- | Convert a list of inline elements to ICML.
-inlinesToICML :: WriterOptions -> Style -> [Inline] -> WS Doc
+inlinesToICML :: PandocMonad m => WriterOptions -> Style -> [Inline] -> WS m Doc
 inlinesToICML opts style lst = vcat `fmap` mapM (inlineToICML opts style) (mergeSpaces lst)
 
 -- | Convert an inline element to ICML.
-inlineToICML :: WriterOptions -> Style -> Inline -> WS Doc
+inlineToICML :: PandocMonad m => WriterOptions -> Style -> Inline -> WS m Doc
 inlineToICML _    style (Str str) = charStyle style $ text $ escapeStringForXML str
 inlineToICML opts style (Emph lst) = inlinesToICML opts (emphName:style) lst
 inlineToICML opts style (Strong lst) = inlinesToICML opts (strongName:style) lst
@@ -433,17 +438,20 @@ inlineToICML opts style SoftBreak =
        WrapPreserve -> charStyle style cr
 inlineToICML _ style LineBreak = charStyle style $ text lineSeparator
 inlineToICML opts style (Math mt str) =
-  cat <$> mapM (inlineToICML opts style) (texMathToInlines mt str)
-inlineToICML _ _ (RawInline f str)
+  lift (texMathToInlines mt str) >>=
+    (fmap cat . mapM (inlineToICML opts style))
+inlineToICML _ _ il@(RawInline f str)
   | f == Format "icml" = return $ text str
-  | otherwise          = return empty
+  | otherwise          = do
+      report $ InlineNotRendered il
+      return empty
 inlineToICML opts style (Link _ lst (url, title)) = do
   content <- inlinesToICML opts (linkName:style) lst
   state $ \st ->
             let ident = if null $ links st
                            then 1::Int
-                           else 1 + (fst $ head $ links st)
-                newst = st{ links = (ident, url):(links st) }
+                           else 1 + fst (head $ links st)
+                newst = st{ links = (ident, url):links st }
                 cont  = inTags True "HyperlinkTextSource"
                          [("Self","htss-"++show ident), ("Name",title), ("Hidden","false")] content
             in  (cont, newst)
@@ -452,9 +460,9 @@ inlineToICML opts style (Note lst) = footnoteToICML opts style lst
 inlineToICML opts style (Span _ lst) = inlinesToICML opts style lst
 
 -- | Convert a list of block elements to an ICML footnote.
-footnoteToICML :: WriterOptions -> Style -> [Block] -> WS Doc
+footnoteToICML :: PandocMonad m => WriterOptions -> Style -> [Block] -> WS m Doc
 footnoteToICML opts style lst =
-  let insertTab (Para ls) = blockToICML opts (footnoteName:style) $ Para $ (Str "\t"):ls
+  let insertTab (Para ls) = blockToICML opts (footnoteName:style) $ Para $ Str "\t":ls
       insertTab block     = blockToICML opts (footnoteName:style) block
   in  do
     contents <- mapM insertTab lst
@@ -466,24 +474,24 @@ footnoteToICML opts style lst =
 
 -- | Auxiliary function to merge Space elements into the adjacent Strs.
 mergeSpaces :: [Inline] -> [Inline]
-mergeSpaces ((Str s):(x:((Str s'):xs))) | isSp x =
+mergeSpaces (Str s:(x:(Str s':xs))) | isSp x =
   mergeSpaces $ Str(s++" "++s') : xs
-mergeSpaces (x:((Str s):xs)) | isSp x = mergeSpaces $ Str (" "++s) : xs
-mergeSpaces ((Str s):(x:xs)) | isSp x = mergeSpaces $ Str (s++" ") : xs
-mergeSpaces (x:xs) = x : (mergeSpaces xs)
+mergeSpaces (x:(Str s:xs)) | isSp x = mergeSpaces $ Str (" "++s) : xs
+mergeSpaces (Str s:(x:xs)) | isSp x = mergeSpaces $ Str (s++" ") : xs
+mergeSpaces (x:xs) = x : mergeSpaces xs
 mergeSpaces []     = []
 
 isSp :: Inline -> Bool
-isSp Space = True
+isSp Space     = True
 isSp SoftBreak = True
-isSp _ = False
+isSp _         = False
 
 -- | Intersperse line breaks
 intersperseBrs :: [Doc] -> Doc
 intersperseBrs = vcat . intersperse (selfClosingTag "Br" []) . filter (not . isEmpty)
 
 -- | Wrap a list of inline elements in an ICML Paragraph Style
-parStyle :: WriterOptions -> Style -> [Inline] -> WS Doc
+parStyle :: PandocMonad m => WriterOptions -> Style -> [Inline] -> WS m Doc
 parStyle opts style lst =
   let slipIn x y = if null y
                       then x
@@ -498,7 +506,7 @@ parStyle opts style lst =
                             begins = filter (isPrefixOf beginsWithName) style
                         in  if null begins
                                then ats
-                               else let i = maybe "" id $ stripPrefix beginsWithName $ head begins
+                               else let i = fromMaybe "" $ stripPrefix beginsWithName $ head begins
                                     in  ("NumberingStartAt", i) : ats
                    else [attrs]
   in  do
@@ -507,16 +515,16 @@ parStyle opts style lst =
       state $ \st -> (cont, st{ blockStyles = Set.insert stlStr $ blockStyles st })
 
 -- | Wrap a Doc in an ICML Character Style.
-charStyle :: Style -> Doc -> WS Doc
+charStyle :: PandocMonad m => Style -> Doc -> WS m Doc
 charStyle style content =
   let (stlStr, attrs) = styleToStrAttr style
       doc = inTags True "CharacterStyleRange" attrs $ inTagsSimple "Content" $ flush content
-  in  do
+  in
       state $ \st ->
-        let styles = if null stlStr
-                        then st
-                        else st{ inlineStyles = Set.insert stlStr $ inlineStyles st }
-        in  (doc, styles)
+    let styles = if null stlStr
+                    then st
+                    else st{ inlineStyles = Set.insert stlStr $ inlineStyles st }
+    in  (doc, styles)
 
 -- | Transform a Style to a tuple of String (eliminating duplicates and ordered) and corresponding attribute.
 styleToStrAttr :: Style -> (String, [(String, String)])
@@ -529,20 +537,18 @@ styleToStrAttr style =
   in  (stlStr, attrs)
 
 -- | Assemble an ICML Image.
-imageICML :: WriterOptions -> Style -> Attr -> Target -> WS Doc
+imageICML :: PandocMonad m => WriterOptions -> Style -> Attr -> Target -> WS m Doc
 imageICML opts style attr (src, _) = do
-  res  <- liftIO $ fetchItem (writerSourceURL opts) src
-  imgS <- case res of
-            Left (_) -> do
-              warn $ "Could not find image `" ++ src ++ "', skipping..."
-              return def
-            Right (img, _) -> do
-              case imageSize img of
+  imgS <- catchError
+          (do (img, _) <- P.fetchItem src
+              case imageSize opts img of
                 Right size -> return size
                 Left msg   -> do
-                  warn $ "Could not determine image size in `" ++
-                    src ++ "': " ++ msg
-                  return def
+                  report $ CouldNotDetermineImageSize src msg
+                  return def)
+           (\e -> do
+               report $ CouldNotFetchResource src (show e)
+               return def)
   let (ow, oh) = sizeInPoints imgS
       (imgWidth, imgHeight) = desiredSizeInPoints opts attr imgS
       hw = showFl $ ow / 2
@@ -571,6 +577,5 @@ imageICML opts style attr (src, _) = do
                    ]
       doc    = inTags True "CharacterStyleRange" attrs
                  $ inTags True "Rectangle" [("Self","uec"), ("StrokeWeight", "0"),
-                     ("ItemTransform", scale++" "++hw++" -"++hh)]
-                 $ (props $$ image)
+                     ("ItemTransform", scale++" "++hw++" -"++hh)] (props $$ image)
   state $ \st -> (doc, st{ inlineStyles = Set.insert stlStr $ inlineStyles st } )
