@@ -1,5 +1,7 @@
+{-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE OverloadedStrings #-}
 {-
-Copyright (C) 2007-2015 John MacFarlane <jgm@berkeley.edu>
+Copyright (C) 2007-2018 John MacFarlane <jgm@berkeley.edu>
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -18,7 +20,7 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 {- |
    Module      : Text.Pandoc.Writers.Man
-   Copyright   : Copyright (C) 2007-2015 John MacFarlane
+   Copyright   : Copyright (C) 2007-2018 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -29,79 +31,96 @@ Conversion of 'Pandoc' documents to groff man page format.
 
 -}
 module Text.Pandoc.Writers.Man ( writeMan) where
-import Text.Pandoc.Definition
-import Text.Pandoc.Templates
-import Text.Pandoc.Shared
-import Text.Pandoc.Writers.Shared
-import Text.Pandoc.Options
-import Text.Pandoc.Readers.TeXMath
-import Text.Printf ( printf )
-import Data.List ( stripPrefix, intersperse, intercalate )
+import Prelude
+import Control.Monad.State.Strict
+import Data.List (intercalate, intersperse, sort, stripPrefix)
+import qualified Data.Map as Map
 import Data.Maybe (fromMaybe)
-import Text.Pandoc.Pretty
+import Data.Text (Text)
+import qualified Data.Text as T
 import Text.Pandoc.Builder (deleteMeta)
-import Control.Monad.State
+import Text.Pandoc.Class (PandocMonad, report)
+import Text.Pandoc.Definition
+import Text.Pandoc.Logging
+import Text.Pandoc.Options
+import Text.Pandoc.Pretty
+import Text.Pandoc.Shared
+import Text.Pandoc.Templates
+import Text.Pandoc.Writers.Math
+import Text.Pandoc.Writers.Shared
+import Text.Printf (printf)
 
 type Notes = [[Block]]
-data WriterState = WriterState { stNotes  :: Notes
-                               , stHasTables :: Bool }
+data WriterState = WriterState { stNotes        :: Notes
+                               , stFontFeatures :: Map.Map Char Bool
+                               , stHasTables    :: Bool }
+
+defaultWriterState :: WriterState
+defaultWriterState = WriterState { stNotes = []
+                                 , stFontFeatures  = Map.fromList [
+                                                       ('I',False)
+                                                     , ('B',False)
+                                                     , ('C',False)
+                                                     ]
+                                 , stHasTables  = False }
 
 -- | Convert Pandoc to Man.
-writeMan :: WriterOptions -> Pandoc -> String
-writeMan opts document = evalState (pandocToMan opts document) (WriterState [] False)
+writeMan :: PandocMonad m => WriterOptions -> Pandoc -> m Text
+writeMan opts document =
+  evalStateT (pandocToMan opts document) defaultWriterState
 
 -- | Return groff man representation of document.
-pandocToMan :: WriterOptions -> Pandoc -> State WriterState String
+pandocToMan :: PandocMonad m => WriterOptions -> Pandoc -> StateT WriterState m Text
 pandocToMan opts (Pandoc meta blocks) = do
   let colwidth = if writerWrapText opts == WrapAuto
                     then Just $ writerColumns opts
                     else Nothing
-  let render' = render colwidth
+  let render' :: Doc -> Text
+      render' = render colwidth
   titleText <- inlineListToMan opts $ docTitle meta
   let title' = render' titleText
   let setFieldsFromTitle =
-       case break (== ' ') title' of
-           (cmdName, rest) -> case break (=='(') cmdName of
-                                   (xs, '(':ys) | not (null ys) &&
-                                                  last ys == ')' ->
+       case T.break (== ' ') title' of
+           (cmdName, rest) -> case T.break (=='(') cmdName of
+                                   (xs, ys) | "(" `T.isPrefixOf` ys
+                                                && ")" `T.isSuffixOf` ys ->
                                      defField "title" xs .
-                                     defField "section" (init ys) .
-                                     case splitBy (=='|') rest of
+                                     defField "section" (T.init $ T.drop 1 ys) .
+                                     case T.splitOn "|" rest of
                                           (ft:hds) ->
-                                            defField "footer" (trim ft) .
+                                            defField "footer" (T.strip ft) .
                                             defField "header"
-                                               (trim $ concat hds)
+                                               (T.strip $ mconcat hds)
                                           [] -> id
                                    _  -> defField "title" title'
   metadata <- metaToJSON opts
-              (fmap (render colwidth) . blockListToMan opts)
-              (fmap (render colwidth) . inlineListToMan opts)
+              (fmap render' . blockListToMan opts)
+              (fmap render' . inlineListToMan opts)
               $ deleteMeta "title" meta
   body <- blockListToMan opts blocks
-  notes <- liftM stNotes get
+  notes <- gets stNotes
   notes' <- notesToMan opts (reverse notes)
   let main = render' $ body $$ notes' $$ text ""
-  hasTables <- liftM stHasTables get
+  hasTables <- gets stHasTables
   let context = defField "body" main
               $ setFieldsFromTitle
               $ defField "has-tables" hasTables
               $ defField "hyphenate" True
-              $ defField "pandoc-version" pandocVersion
-              $ metadata
+              $ defField "pandoc-version" pandocVersion metadata
   case writerTemplate opts of
        Nothing  -> return main
-       Just tpl -> return $ renderTemplate' tpl context
+       Just tpl -> renderTemplate' tpl context
 
 -- | Return man representation of notes.
-notesToMan :: WriterOptions -> [[Block]] -> State WriterState Doc
+notesToMan :: PandocMonad m => WriterOptions -> [[Block]] -> StateT WriterState m Doc
 notesToMan opts notes =
   if null notes
      then return empty
-     else mapM (\(num, note) -> noteToMan opts num note) (zip [1..] notes) >>=
+     else zipWithM (noteToMan opts) [1..] notes >>=
           return . (text ".SH NOTES" $$) . vcat
 
 -- | Return man representation of a note.
-noteToMan :: WriterOptions -> Int -> [Block] -> State WriterState Doc
+noteToMan :: PandocMonad m => WriterOptions -> Int -> [Block] -> StateT WriterState m Doc
 noteToMan opts num note = do
   contents <- blockListToMan opts note
   let marker = cr <> text ".SS " <> brackets (text (show num))
@@ -123,11 +142,11 @@ escapeString = escapeStringUsing manEscapes
 
 -- | Escape a literal (code) section for Man.
 escapeCode :: String -> String
-escapeCode = concat . intersperse "\n" . map escapeLine . lines  where
+escapeCode = intercalate "\n" . map escapeLine . lines  where
   escapeLine codeline =
     case escapeStringUsing (manEscapes ++ backslashEscapes "\t ") codeline of
       a@('.':_) -> "\\&" ++ a
-      b       -> b
+      b         -> b
 
 -- We split inline lists into sentences, and print one sentence per
 -- line.  groff/troff treats the line-ending period differently.
@@ -139,8 +158,8 @@ breakSentence [] = ([],[])
 breakSentence xs =
   let isSentenceEndInline (Str ys@(_:_)) | last ys == '.' = True
       isSentenceEndInline (Str ys@(_:_)) | last ys == '?' = True
-      isSentenceEndInline (LineBreak) = True
-      isSentenceEndInline _         = False
+      isSentenceEndInline LineBreak      = True
+      isSentenceEndInline _              = False
       (as, bs) = break isSentenceEndInline xs
   in  case bs of
            []             -> (as, [])
@@ -160,9 +179,10 @@ splitSentences xs =
   in  if null rest then [sent] else sent : splitSentences rest
 
 -- | Convert Pandoc block element to man.
-blockToMan :: WriterOptions -- ^ Options
-                -> Block         -- ^ Block element
-                -> State WriterState Doc
+blockToMan :: PandocMonad m
+           => WriterOptions -- ^ Options
+           -> Block         -- ^ Block element
+           -> StateT WriterState m Doc
 blockToMan _ Null = return empty
 blockToMan opts (Div _ bs) = blockListToMan opts bs
 blockToMan opts (Plain inlines) =
@@ -173,9 +193,11 @@ blockToMan opts (Para inlines) = do
   return $ text ".PP" $$ contents
 blockToMan opts (LineBlock lns) =
   blockToMan opts $ linesToPara lns
-blockToMan _ (RawBlock f str)
+blockToMan _ b@(RawBlock f str)
   | f == Format "man" = return $ text str
-  | otherwise         = return empty
+  | otherwise         = do
+      report $ BlockNotRendered b
+      return empty
 blockToMan _ HorizontalRule = return $ text ".PP" $$ text "   *   *   *   *   *"
 blockToMan opts (Header level _ inlines) = do
   contents <- inlineListToMan opts inlines
@@ -205,12 +227,12 @@ blockToMan opts (Table caption alignments widths headers rows) =
                    then repeat ""
                    else map (printf "w(%0.1fn)" . (70 *)) widths
   -- 78n default width - 8n indent = 70n
-  let coldescriptions = text $ intercalate " "
+  let coldescriptions = text $ unwords
                         (zipWith (\align width -> aligncode align ++ width)
                         alignments iwidths) ++ "."
   colheadings <- mapM (blockListToMan opts) headers
   let makeRow cols = text "T{" $$
-                     (vcat $ intersperse (text "T}@T{") cols) $$
+                     vcat (intersperse (text "T}@T{") cols) $$
                      text "T}"
   let colheadings' = if all null headers
                         then empty
@@ -227,7 +249,8 @@ blockToMan opts (BulletList items) = do
   return (vcat contents)
 blockToMan opts (OrderedList attribs items) = do
   let markers = take (length items) $ orderedListMarkers attribs
-  let indent = 1 + (maximum $ map length markers)
+  let indent = 1 +
+                     maximum (map length markers)
   contents <- mapM (\(num, item) -> orderedListItemToMan opts num indent item) $
               zip markers items
   return (vcat contents)
@@ -236,11 +259,11 @@ blockToMan opts (DefinitionList items) = do
   return (vcat contents)
 
 -- | Convert bullet list item (list of blocks) to man.
-bulletListItemToMan :: WriterOptions -> [Block] -> State WriterState Doc
+bulletListItemToMan :: PandocMonad m => WriterOptions -> [Block] -> StateT WriterState m Doc
 bulletListItemToMan _ [] = return empty
-bulletListItemToMan opts ((Para first):rest) =
-  bulletListItemToMan opts ((Plain first):rest)
-bulletListItemToMan opts ((Plain first):rest) = do
+bulletListItemToMan opts (Para first:rest) =
+  bulletListItemToMan opts (Plain first:rest)
+bulletListItemToMan opts (Plain first:rest) = do
   first' <- blockToMan opts (Plain first)
   rest' <- blockListToMan opts rest
   let first'' = text ".IP \\[bu] 2" $$ first'
@@ -254,14 +277,15 @@ bulletListItemToMan opts (first:rest) = do
   return $ text "\\[bu] .RS 2" $$ first' $$ rest' $$ text ".RE"
 
 -- | Convert ordered list item (a list of blocks) to man.
-orderedListItemToMan :: WriterOptions -- ^ options
-                          -> String   -- ^ order marker for list item
-                          -> Int      -- ^ number of spaces to indent
-                          -> [Block]  -- ^ list item (list of blocks)
-                          -> State WriterState Doc
+orderedListItemToMan :: PandocMonad m
+                     => WriterOptions -- ^ options
+                     -> String   -- ^ order marker for list item
+                     -> Int      -- ^ number of spaces to indent
+                     -> [Block]  -- ^ list item (list of blocks)
+                     -> StateT WriterState m Doc
 orderedListItemToMan _ _ _ [] = return empty
-orderedListItemToMan opts num indent ((Para first):rest) =
-  orderedListItemToMan opts num indent ((Plain first):rest)
+orderedListItemToMan opts num indent (Para first:rest) =
+  orderedListItemToMan opts num indent (Plain first:rest)
 orderedListItemToMan opts num indent (first:rest) = do
   first' <- blockToMan opts first
   rest' <- blockListToMan opts rest
@@ -273,44 +297,47 @@ orderedListItemToMan opts num indent (first:rest) = do
   return $ first'' $$ rest''
 
 -- | Convert definition list item (label, list of blocks) to man.
-definitionListItemToMan :: WriterOptions
-                             -> ([Inline],[[Block]])
-                             -> State WriterState Doc
+definitionListItemToMan :: PandocMonad m
+                        => WriterOptions
+                        -> ([Inline],[[Block]])
+                        -> StateT WriterState m Doc
 definitionListItemToMan opts (label, defs) = do
   labelText <- inlineListToMan opts label
   contents <- if null defs
                  then return empty
-                 else liftM vcat $ forM defs $ \blocks -> do
-                        let (first, rest) = case blocks of
-                              ((Para x):y) -> (Plain x,y)
-                              (x:y)        -> (x,y)
-                              []           -> error "blocks is null"
-                        rest' <- liftM vcat $
-                                  mapM (\item -> blockToMan opts item) rest
-                        first' <- blockToMan opts first
-                        return $ first' $$ text ".RS" $$ rest' $$ text ".RE"
+                 else liftM vcat $ forM defs $ \blocks ->
+                        case blocks of
+                          (x:xs) -> do
+                            first' <- blockToMan opts $
+                                      case x of
+                                           Para y -> Plain y
+                                           _      -> x
+                            rest' <- liftM vcat $ mapM
+                                        (\item -> blockToMan opts item) xs
+                            return $ first' $$
+                                     text ".RS" $$ rest' $$ text ".RE"
+                          [] -> return empty
   return $ text ".TP" $$ nowrap (text ".B " <> labelText) $$ contents
 
 -- | Convert list of Pandoc block elements to man.
-blockListToMan :: WriterOptions -- ^ Options
-                    -> [Block]       -- ^ List of block elements
-                    -> State WriterState Doc
+blockListToMan :: PandocMonad m
+               => WriterOptions -- ^ Options
+               -> [Block]       -- ^ List of block elements
+               -> StateT WriterState m Doc
 blockListToMan opts blocks =
   mapM (blockToMan opts) blocks >>= (return . vcat)
 
 -- | Convert list of Pandoc inline elements to man.
-inlineListToMan :: WriterOptions -> [Inline] -> State WriterState Doc
+inlineListToMan :: PandocMonad m => WriterOptions -> [Inline] -> StateT WriterState m Doc
 inlineListToMan opts lst = mapM (inlineToMan opts) lst >>= (return . hcat)
 
 -- | Convert Pandoc inline element to man.
-inlineToMan :: WriterOptions -> Inline -> State WriterState Doc
+inlineToMan :: PandocMonad m => WriterOptions -> Inline -> StateT WriterState m Doc
 inlineToMan opts (Span _ ils) = inlineListToMan opts ils
-inlineToMan opts (Emph lst) = do
-  contents <- inlineListToMan opts lst
-  return $ text "\\f[I]" <> contents <> text "\\f[]"
-inlineToMan opts (Strong lst) = do
-  contents <- inlineListToMan opts lst
-  return $ text "\\f[B]" <> contents <> text "\\f[]"
+inlineToMan opts (Emph lst) =
+  withFontFeature 'I' (inlineListToMan opts lst)
+inlineToMan opts (Strong lst) =
+  withFontFeature 'B' (inlineListToMan opts lst)
 inlineToMan opts (Strikeout lst) = do
   contents <- inlineListToMan opts lst
   return $ text "[STRIKEOUT:" <> contents <> char ']'
@@ -330,22 +357,26 @@ inlineToMan opts (Quoted DoubleQuote lst) = do
 inlineToMan opts (Cite _ lst) =
   inlineListToMan opts lst
 inlineToMan _ (Code _ str) =
-  return $ text $ "\\f[C]" ++ escapeCode str ++ "\\f[]"
+  withFontFeature 'C' (return (text $ escapeCode str))
 inlineToMan _ (Str str@('.':_)) =
   return $ afterBreak "\\&" <> text (escapeString str)
 inlineToMan _ (Str str) = return $ text $ escapeString str
 inlineToMan opts (Math InlineMath str) =
-  inlineListToMan opts $ texMathToInlines InlineMath str
+  lift (texMathToInlines InlineMath str) >>= inlineListToMan opts
 inlineToMan opts (Math DisplayMath str) = do
-  contents <- inlineListToMan opts $ texMathToInlines DisplayMath str
+  contents <- lift (texMathToInlines DisplayMath str) >>= inlineListToMan opts
   return $ cr <> text ".RS" $$ contents $$ text ".RE"
-inlineToMan _ (RawInline f str)
+inlineToMan _ il@(RawInline f str)
   | f == Format "man" = return $ text str
-  | otherwise         = return empty
-inlineToMan _ (LineBreak) = return $
+  | otherwise         = do
+      report $ InlineNotRendered il
+      return empty
+inlineToMan _ LineBreak = return $
   cr <> text ".PD 0" $$ text ".P" $$ text ".PD" <> cr
 inlineToMan _ SoftBreak = return space
 inlineToMan _ Space = return space
+inlineToMan opts (Link _ txt ('#':_, _)) =
+  inlineListToMan opts txt -- skip internal links
 inlineToMan opts (Link _ txt (src, _)) = do
   linktext <- inlineListToMan opts txt
   let srcSuffix = fromMaybe src (stripPrefix "mailto:" src)
@@ -355,7 +386,7 @@ inlineToMan opts (Link _ txt (src, _)) = do
                                  char '<' <> text srcSuffix <> char '>'
            _                  -> linktext <> text " (" <> text src <> char ')'
 inlineToMan opts (Image attr alternate (source, tit)) = do
-  let txt = if (null alternate) || (alternate == [Str ""]) ||
+  let txt = if null alternate || (alternate == [Str ""]) ||
                (alternate == [Str source]) -- to prevent autolinks
                then [Str "image"]
                else alternate
@@ -364,6 +395,24 @@ inlineToMan opts (Image attr alternate (source, tit)) = do
 inlineToMan _ (Note contents) = do
   -- add to notes in state
   modify $ \st -> st{ stNotes = contents : stNotes st }
-  notes <- liftM stNotes get
-  let ref = show $ (length notes)
+  notes <- gets stNotes
+  let ref = show (length notes)
   return $ char '[' <> text ref <> char ']'
+
+fontChange :: PandocMonad m => StateT WriterState m Doc
+fontChange = do
+  features <- gets stFontFeatures
+  let filling = sort [c | (c,True) <- Map.toList features]
+  return $ text $ "\\f[" ++ filling ++ "]"
+
+withFontFeature :: PandocMonad m
+                => Char
+                -> StateT WriterState m Doc
+                -> StateT WriterState m Doc
+withFontFeature c action = do
+  modify $ \st -> st{ stFontFeatures = Map.adjust not c $ stFontFeatures st }
+  begin <- fontChange
+  d <- action
+  modify $ \st -> st{ stFontFeatures = Map.adjust not c $ stFontFeatures st }
+  end <- fontChange
+  return $ begin <> d <> end
